@@ -28,11 +28,15 @@
 #'   processes the output from the evaluation. The default simply prints the
 #'   visible return values.
 #' @param filename string overrriding the \code{\link[base]{srcfile}} filename.
+#' @param include_timing if \code{TRUE}, evaluate will wrap each input
+#'   expression in \code{system.time()}, which will be accessed by following
+#'   \code{replay()} call to produce timing information for each evaluated
+#'   command.
 #' @import graphics grDevices stringr utils
 evaluate <- function(input, envir = parent.frame(), enclos = NULL, debug = FALSE,
                      stop_on_error = 0L, keep_warning = TRUE, keep_message = TRUE,
                      new_device = TRUE, output_handler = default_output_handler,
-                     filename = NULL) {
+                     filename = NULL, include_timing = FALSE) {
   stop_on_error <- as.integer(stop_on_error)
   stopifnot(length(stop_on_error) == 1)
 
@@ -71,7 +75,8 @@ evaluate <- function(input, envir = parent.frame(), enclos = NULL, debug = FALSE
       envir = envir, enclos = enclos, debug = debug, last = i == length(out),
       use_try = stop_on_error != 2L,
       keep_warning = keep_warning, keep_message = keep_message,
-      output_handler = output_handler)
+      output_handler = output_handler,
+      include_timing = include_timing)
 
     if (stop_on_error > 0L) {
       errs <- vapply(out[[i]], is.error, logical(1))
@@ -88,7 +93,7 @@ evaluate_call <- function(call, src = NULL,
                           envir = parent.frame(), enclos = NULL,
                           debug = FALSE, last = FALSE, use_try = FALSE,
                           keep_warning = TRUE, keep_message = TRUE,
-                          output_handler = new_output_handler()) {
+                          output_handler = new_output_handler(), include_timing = FALSE) {
   if (debug) message(src)
 
   if (is.null(call) && !last) {
@@ -114,6 +119,11 @@ evaluate_call <- function(call, src = NULL,
     output <<- c(output, out)
   }
 
+  flush_old <- .env$flush_console; on.exit({
+    .env$flush_console <- flush_old
+  }, add = TRUE)
+  .env$flush_console <- function() handle_output(FALSE)
+
   # Hooks to capture plot creation
   capture_plot <- function() {
     handle_output(TRUE)
@@ -131,6 +141,9 @@ evaluate_call <- function(call, src = NULL,
 
   # Handlers for warnings, errors and messages
   wHandler <- if (keep_warning) function(wn) {
+    # do not handle the warning as it will be raised as error after
+    if (getOption("warn") >= 2) return()
+
     if (getOption("warn") >= 0) {
       handle_condition(wn)
       output_handler$warning(wn)
@@ -155,13 +168,30 @@ evaluate_call <- function(call, src = NULL,
     handle <- force
   }
   value_handler <- output_handler$value
+  if (include_timing) {
+    timing_fn <- function(x) system.time(x)[1:3]
+  } else {
+    timing_fn <- function(x) {x; NULL};
+  }
+
+  if (length(funs <- .env$inject_funs)) {
+    funs_names <- names(funs)
+    funs_new <- !vapply(funs_names, exists, logical(1), envir, inherits = FALSE)
+    funs_names <- funs_names[funs_new]
+    funs <- funs[funs_new]
+    on.exit(rm(list = funs_names, envir = envir), add = TRUE)
+    for (i in seq_along(funs_names)) assign(funs_names[i], funs[[i]], envir)
+  }
+
   multi_args <- length(formals(value_handler)) > 1
   for (expr in call) {
-    handle(ev <- withCallingHandlers(
+    srcindex <- length(output)
+    time <- timing_fn(handle(ev <- withCallingHandlers(
       withVisible(eval(expr, envir, enclos)),
-      warning = wHandler, error = eHandler, message = mHandler))
+      warning = wHandler, error = eHandler, message = mHandler)))
     handle_output(TRUE)
-
+    if (!is.null(time))
+      attr(output[[srcindex]]$src, 'timing') <- time
 
     # If visible or the value handler has multi args, process and capture output
     if (ev$visible || multi_args) {
@@ -185,3 +215,30 @@ evaluate_call <- function(call, src = NULL,
   output
 }
 
+#' Inject functions into the environment of \code{evaluate()}
+#'
+#' Create functions in the environment specified in the \code{envir} argument of
+#' \code{evaluate()}. This can be helpful if you want to substitute certain
+#' functions when evaluating the code. To make sure it does not wipe out
+#' existing functions in the environment, only functions that do not exist in
+#' the environment are injected.
+#' @param ... Named arguments of functions. If empty, previously injected
+#'   functions will be emptied.
+#' @note For expert use only. Do not use it unless you clearly understand it.
+#' @keywords internal
+#' @examples library(evaluate)
+#' # normally you cannot capture the output of system
+#' evaluate("system('R --version')")
+#'
+#' # replace the system() function
+#' inject_funs(system = function(...) cat(base::system(..., intern = TRUE), sep = '\n'))
+#'
+#' evaluate("system('R --version')")
+#'
+#' inject_funs()  # empty previously injected functions
+#' @export
+inject_funs <- function(...) {
+  funs <- list(...)
+  funs <- funs[names(funs) != '']
+  .env$inject_funs <- Filter(is.function, funs)
+}
